/* 
 * Created : Sat Dec 21 00:43:53 CET 2002
 * Updated : Mon Jan 27 14:24:54 CET 2003
 *
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include "include/kernsh.h"
#include "libdisasm/libdis.h"

/* adapted from quikdis.c in libdisasm */
int
disasm( size_t start_address, int len )
{
	Elf32_Ehdr * target_header;
	Elf32_Phdr * program_header;
	struct stat tmpstat;
	struct instr curr_inst;
	void * image_vmlinux;
	unsigned char * buf;
	unsigned char symbol_name[ 2*sizeof(size_t)+2 ];
	size_t p_symbol;
	int entry;
	int ftarget;
	int n;
	int x;
	int offset;
	int i;
	int j;
	int size;
	int c;

	disassemble_init( 0, INTEL_SYNTAX );

	/* read ELF header */
	ftarget = found_vmlinux();
	if ( ftarget < 0 ) {
		perror( "open()" );
		return( -1 );
	}

	n = fstat( ftarget, &tmpstat );
	if ( n < 0 ) {
		perror( "fstat()" );
		return( -1 );
	}

	image_vmlinux = mmap( 0, tmpstat.st_size, PROT_READ, MAP_SHARED, 
	ftarget, 0 );
	if ( (int)image_vmlinux < 1 ) {
		perror( "mmap()" );
		return( -1 );
	}
	close( ftarget );

	target_header = image_vmlinux;

	/* read program header table */
	for ( x = 0; x < target_header->e_phnum; x++ ) {
		program_header = image_vmlinux + target_header->e_phoff + 
		( x * target_header->e_phentsize );

		/* If entry point is in this section */
		if ( target_header->e_entry >= program_header->p_vaddr && 
		target_header->e_entry <= 
		(program_header->p_vaddr + program_header->p_filesz)) {

			/* resolve entry point RVA to a file offset */
			entry = target_header->e_entry - 
			( program_header->p_vaddr - program_header->p_offset );

			/* read entire program segment into buffer */
			offset = start_address - target_header->e_entry;
			buf = image_vmlinux + entry + offset;

			i = 0;
			while ( i < len ) {
				memset( &curr_inst, 0, sizeof(struct code) );
				fprintf( stdout, "%#x\t", 
				target_header->e_entry+offset+i );

				size = disassemble_address( buf+i, &curr_inst );
				if ( size ) {
					for ( n = 0; n < 12; n++ ) {
						if ( n < size ) {
							fprintf( stdout, "%02x ", buf[i+n] );
						}
						else {
							fputs( "   ", stdout );
						}
					}
					fprintf( stdout, "%s", curr_inst.mnemonic );
					if ( curr_inst.dest[0] != 0 ) {
						if ( strstr(curr_inst.mnemonic, "call") ) {
							p_symbol = 
							( target_header->e_entry + offset + i ) + 
							strtoul( curr_inst.dest, NULL, 16 );

							snprintf( symbol_name, sizeof(symbol_name), 
							"%#x", p_symbol );
							fprintf( stdout, "\t%s", 
							get_symbol_by_addr(symbol_name) );
						}
						else {
							fprintf( stdout, "\t%s", curr_inst.dest );
						}
					}
					if ( curr_inst.src[0] != 0 ) {
						fprintf( stdout, ", %s", curr_inst.src );
					}
					if ( curr_inst.aux[0] != 0 ) {
						fprintf( stdout, ", %s", curr_inst.aux );
					}
					fputs( "\n", stdout );
					i += size;
				} 
				else {
					fatal( "invalid opcode %02x\n", buf[i] );
					i++;
				}

				j++;
				if ( !((j+1)%60) ) {
					c = enter();
					if ( c == '\r' || c == '\n' ) {
						continue;
					}
					if ( c == 'q' ) {
						getchar();
						return( 0 );
					}
				}
			}
		}
	}

	munmap( image_vmlinux, tmpstat.st_size );
	disassemble_cleanup();
	return( 0 );
}
