/* 
 * Created : Sat Dec 21 00:43:53 CET 2002
 * Updated : Thu Jan 30 01:16:11 CET 2003
 *
 * Copyright (C) 2002-2003 sauron@kernsh.org
 * Copyright (C) 2002-2003 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 * 
 * The program kernsh is (C) 2002-2003 sauron/zorgon. This program is 
 * free software; you can redistribute it and/or modify it under the 
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions. 
 * 
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.                                                                  
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for 
 * possible incorporation into the main distribution. By sending these 
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that 
 * you are offering them the unlimited, non-exclusive right to reuse, 
 * modify, and relicense the code. This is important because the 
 * inability to relicense code has caused devastating problems for other 
 * Free Software projects (such as KDE and NASM). Kernsh will always be 
 * available Open Source. If you wish to specify special license 
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include "include/kernsh.h"

#define K_SYMBOLS 1 /* want info about symbols */
#define K_INFO 2 /* want extended module info */
#define K_REFS 4 /* want info about references */
#define NEW_MOD_DELETED 2

/* from modstat.h in modutils-2.4.15 */
struct module_stat {
    char * name;
    unsigned long addr;
    unsigned long modstruct; /* COMPAT_2_0! */
							 /* depends on architecture? */
    unsigned long size;
    unsigned long flags;
    long usecount;
    size_t nsyms;
    struct module_symbol * syms;
    size_t nrefs;
    struct module_stat ** refs;
    unsigned long status;
};

struct module_stat * module_stat;
size_t n_module_stat;
char * module_name_list;
size_t l_module_name_list;
struct module_symbol * ksyms;
size_t nksyms;
int k_new_syscalls;
static int sym_index = -1;
static int force = 1;
static int sym_num = -1;
int modules_not_supported;

int 
isprintable( char * str )
{
	if ( *str == '\0' ) {
		return( 0 );
	}

	while ( *str ) {
		if ( !((*str >= 32) && (*str <= 126)) ) {
			return( 0 );
		}
		str++;
	}
	return( 1 );
}

int
probepath( int fd, size_t addr, struct module mod )
{
	char temp[ mod.size ];
	unsigned long * a;
	int i; 
	int j; 
	int n;

	a = malloc( BUFSIZ*sizeof(char) );
	if ( a == NULL ) {
		perror( "malloc()" );
		return( -1 );
	}

	n = read_kmem( &temp, addr, mod.size, fd );
	if ( n < 0 ) {
		return( -1 );
	}

	for ( i = 0; i != mod.size; i++ ) {
		if ( temp[i] == '/' ) {
			for ( j = i ; j < i + 70 ; j++ ) {
				if ( temp [j] == '.' ) {
					if ( temp [++j] == 'o' ) {
						for ( ; i != j+1; i++ ) {
							fprintf( stdout, "%c", temp[i] );
						}
					}
				}
			}
		}
	}
	return( 0 );
}

void 
drop( void )
{
	/* Clean the slate for multiple runs */
	if ( module_stat ) {
		struct module_stat * m;
		int i;

		for ( i = 0, m = module_stat; i < n_module_stat; 
		++i, ++m ) {
			if ( m->syms ) {
				free( m->syms );
			}
			
			if ( m->refs ) {
				free( m->refs );
			}
			
			free( module_stat );
			module_stat = NULL;
			n_module_stat = 0;
		}
	}
		
	if ( module_name_list ) {
		free( module_name_list );
		module_name_list = NULL;
		l_module_name_list = 0;
	}

	if ( ksyms ) {
		free( ksyms );
		ksyms = NULL;
		nksyms = 0;
	}
	return;
}

int 
get_kernel_info( int type )
{
	struct module_stat * modules;
	struct module_stat * m;
	struct module_symbol * syms;
	struct module_symbol * s;
	size_t ret;
	size_t bufsize;
	size_t nmod;
	size_t nsyms;
	size_t i;
	size_t j;
	char * module_names;
	char * mn;

	drop();

	/* Collect the loaded modules */
	module_names = (char *)malloc( BUFSIZ );
	if ( module_names == NULL ) {
		perror( "malloc()" );
		return( -1 );
	}

	while ( query_module(NULL, QM_MODULES, module_names, 
	bufsize, &ret) ) {
		if ( errno != ENOSPC ) {
			fatal( "QM_MODULES: %m\n" );
			return( -1 );
		}
		module_names = ( char * )realloc( module_names, 
		bufsize = ret );
	}

	module_name_list = module_names;
	l_module_name_list = bufsize;
	n_module_stat = nmod = ret;
	module_stat = modules = 
	( void * )malloc( nmod * sizeof(struct module_stat) );
	if ( module_stat == NULL ) {
		perror( "malloc()" );
		return( -1 );
	}
	memset( modules, 0, nmod * sizeof(struct module_stat) );

	/* Collect the info from the modules */
	for ( i = 0, mn = module_names, m = modules;
         i < nmod;
         ++i, ++m, mn += strlen(mn) + 1 ) {
        struct module_info info;

        m->name = mn;
		if ( query_module(mn, QM_INFO, &info, sizeof(info), &ret) ) {
            if ( errno == ENOENT ) {
            /* The module was removed out from underneath us. */
                m->flags = NEW_MOD_DELETED;
                continue;
            }
            /* else oops */
            fatal( "module %s: QM_INFO: %m", mn );
            return( -1 );
        }

        m->addr = info.addr;

        if ( type & K_INFO ) {
            m->size = info.size;
            m->flags = info.flags;
            m->usecount = info.usecount;
            m->modstruct = info.addr;
        }
	
		/* Want info about symbols */
        if ( type & K_SYMBOLS ) { 
		syms = ( void * )malloc( bufsize = BUFSIZ );
		while ( query_module(mn, QM_SYMBOLS, syms, 
		bufsize, &ret) ) {
			if ( errno == ENOSPC ) {
				syms = ( void * )realloc( syms, 
				bufsize = ret );
				continue;
			}	
			if ( errno == ENOENT ) {
				/* The module was removed out
   				 * from underneath us.
                 */
   				m->flags = NEW_MOD_DELETED;
   				free( syms );
   				goto next;
   			} 
			else {
				fatal( "module %s: QM_SYMBOLS: %m", mn );
				return( -1 );
			}
		}
		nsyms = ret;
	
		m->nsyms = nsyms;
		m->syms = syms;

		/* Convert string offsets to string pointers */
		for ( j = 0, s = syms; j < nsyms; ++j, ++s ) {
			s->name += ( size_t )syms;
		}
	}
	next:
	}

	/* Want info about symbols */
	if ( type & K_SYMBOLS ) { 
		/* Collect the kernel's symbols. */
		syms = ( void * )malloc( bufsize = 16 * 1024 );
		while ( query_module(NULL, QM_SYMBOLS, syms, 
		bufsize, &ret) ) {
			if ( errno != ENOSPC ) {
				fatal( "kernel: QM_SYMBOLS: %m" );
				return( -1 );
			}
			syms = ( void * )realloc( syms, bufsize = ret );
		}
		nksyms = nsyms = ret;
		ksyms = syms;
	
		/* Convert string offsets to string pointers */
		for ( j = 0, s = syms; j < nsyms; ++j, ++s ) {
			s->name += ( size_t )syms;
		}
	}
	return( 0 );
}
/* modutils-2.4.15 */

size_t 
find_module_addr( void )
{
	char * module_names; 
	char * mname;
	size_t bufsize; 
	size_t ret;
	struct module_info info;
	int n;

	module_names = ( char * )malloc( BUFSIZ );
	if ( module_names == NULL ) {
		perror( "malloc()" );
		return( -1 );
	}

retry_kern_symbol_load:
	n = query_module( NULL, QM_MODULES, module_names, 
	bufsize, &ret );
	if ( n ) {
		if ( errno == ENOSPC ) {
			module_names = ( char * )realloc( module_names, 
			bufsize = ret );
			goto retry_kern_symbol_load;
		}
		fatal( "find_module_addr: QM_MODULES error %d\n", errno );
		return( -1 );
	}

	mname = module_names;
	free( module_names );

	n = query_module( mname, QM_INFO, &info, sizeof(info), &ret );
	if ( n ) {
		fatal( "find_module_addr: QM_INFO error %d\n", errno );
		return( -1 );
	}
	return( info.addr );
}

Elf32_Sym * 
local_sym_find( Elf32_Sym * symtab, int n, char * strtab, 
const char * name )
{
	int i;

	for ( i = 0; i < n; i++ ) {
		if ( !strcmp(&strtab[symtab[i].st_name], name) ) {
			return( &symtab[i] );
		}
	}
	return( 0 );
}

Elf32_Sym * 
localall_sym_find( module_t * module, const char * name ) 
{
	Elf32_Sym * sym;
    char * strtab = module->section[module->strtabidx]; 
    int i;

    for ( i = 0; i < module->ehdr.e_shnum; i++ ) {
		Elf32_Shdr * shdr = &module->shdr[i];
		if ( shdr->sh_type == SHT_SYMTAB ) {
			sym = local_sym_find( (Elf32_Sym *)module->section[i],
			shdr->sh_size/sizeof(Elf32_Sym),
			strtab,
			name );
			if ( sym ) {
				return( sym );
			}
		}
	}
	return( 0 );
}

int     
check_elf_file( Elf32_Ehdr * ehdr )
{
	if ( strncmp(ehdr->e_ident, ELFMAG, SELFMAG) ) {
		return( -1 );
	}
	if ( ehdr->e_type != ET_REL ) {
		return( -1 );
	}
	if ( ehdr->e_machine != EM_386 && ehdr->e_machine != EM_486 ) {
		return( -1 );
	}
	if ( ehdr->e_version != EV_CURRENT ) {
		return( -1 );
	}
	return( 0 );
}

int     
load_section( char ** p, int fd, Elf32_Shdr * shdr )
{
	int n;
	
    if ( !(*p = (char *)xmalloc(shdr->sh_size)) ) {
		fatal( "not enough memory !\n" );
		return( -1 );
	}

	n = read_kmem( *p, shdr->sh_offset, shdr->sh_size, fd );
    if ( n < 0 ) {
		fatal( "cannot read section !\n" );
		return( -1 );
	}
    return( 0 );
}

int 
load_module( module_t * module, int fd )
{
	Elf32_Ehdr * ehdr = &module->ehdr;
	Elf32_Shdr * shdr;
	char ** sectionp;
	int slen;
	int i;
	int n;
	
	n = read_kmem( ehdr, 0, sizeof(Elf32_Ehdr), fd );
    if ( n < 0 ) {
		fatal( "cannot read ELF header of module !\n" );
		return( -1 );
	}

	/* ELF checks */
	n = check_elf_file( ehdr );
    if ( n < 0 ) {
        fatal( "File not ELF !\n" );
		return( -1 );
	}

    slen = sizeof(Elf32_Shdr) * ehdr->e_shnum;
    if ( (module->shdr=(Elf32_Shdr *)xmalloc(slen)) <= 0 ) {
		fatal( "not enough memory !\n" );
		return( -1 );
	}

    if ( (module->section= 
	(char **)xmalloc(sizeof(char **)*ehdr->e_shnum)) <= 0 ) {
		fatal( "not enough memory !\n" );
		return( -1 );
	}

    n = read_kmem( module->shdr, ehdr->e_shoff, slen, fd );
    if ( n < 0 ) {
		return( -1 );
	}

    for ( i = 0, sectionp = module->section, shdr = module->shdr; 
	i < ehdr->e_shnum; i++, sectionp++ ) {
		switch ( shdr->sh_type ) {
			case SHT_NULL:
			case SHT_NOTE:
			case SHT_NOBITS:
			break;

			case SHT_STRTAB:
				n = load_section( sectionp, fd, shdr );
				if ( n < 0 ) {
					return( -1 );
				}
				if ( i != ehdr->e_shstrndx ) {
					module->strtabidx = i;
				}
				break;

			case SHT_SYMTAB:
				sym_index = i;
				sym_num = shdr->sh_size / shdr->sh_entsize;

			case SHT_PROGBITS:
			case SHT_REL:
				n = load_section( sectionp, fd, shdr );
				if ( n < 0 ) {
					return( -1 );
				}
				break;

			default:
				fatal( "no handler for section (type): %i\n", 
				shdr->sh_type );
				return( -1 );
		}
		++shdr;
	}

	if ( sym_index == -1 || sym_num == -1 || !module->strtabidx ) {
		fatal( "symbol table or string table not found !\n" );
		return( -1 );
	}
	return( 0 );
}

int 
relocate( module_t * module, Elf32_Rel * rel, Elf32_Shdr * shdr )
{
	Elf32_Sym * symtab = (Elf32_Sym *)module->section[shdr->sh_link];
	Elf32_Sym * sym = &symtab[ELF32_R_SYM(rel->r_info)];
	Elf32_Addr addr;
	Elf32_Shdr * targshdr = &module->shdr[shdr->sh_info];
	Elf32_Addr dot =  targshdr->sh_addr + rel->r_offset;
	Elf32_Addr * loc = 
	(Elf32_Addr *)(module->section[shdr->sh_info] + rel->r_offset);
    char * name = &module->section[module->strtabidx][sym->st_name];

    if ( ELF32_ST_BIND(sym->st_info) != STB_LOCAL ) {
		unsigned long ksym;
		if ( force ) {
			char novname[60];
			int len;

			len = strlen( name );
			if ( len > 10 && !strncmp(name+len-10, "_R", 2) ) {
				strncpy( novname, name, len-10 );
				novname[len-10] = 0;
				ksym = get_symbol_by_name( novname );
			}
			else {
				ksym = get_symbol_by_name( name );
			}
		}
		else {
			ksym = get_symbol_by_name( name );
		}

		if ( ksym ) {
			addr = ksym;
			goto next;
		}

		if ( sym->st_shndx == 0 || 
		sym->st_shndx > module->ehdr.e_shnum ) {
			if ( sym->st_shndx == SHN_COMMON ) {
				*loc = sym->st_value;
				return( 0 );
			}
			fatal( "undefined symbol (%s)\n", name );
			return( -1 );
		} 
	}

    addr = sym->st_value + module->shdr[sym->st_shndx].sh_addr;

next:
	if ( targshdr->sh_type == SHT_SYMTAB ) {
		return( 0 );
	}

	if ( targshdr->sh_type != SHT_PROGBITS ) {
		fatal( "Rel not PROGBITS or SYMTAB (type: %i)\n", 
		targshdr->sh_type );
		return( -1 );
	}

	switch ( ELF32_R_TYPE(rel->r_info) ) {
		case R_386_NONE:
		break;

		case R_386_PLT32:
		case R_386_PC32:
			*loc -= dot;	/* *loc += addr - dot	*/

		case R_386_32:
			*loc += addr;
			break;

		default:
			fatal( "no handler for relocation (type): %i", 
			ELF32_R_TYPE(rel->r_info) );
			return( -1 );
	}
	return( 0 );
}

int 
relocate_module( module_t * module )
{
    int i;
    int j;
	int n;

	for ( i = 0; i < module->ehdr.e_shnum; i++ ) {
		if ( module->shdr[i].sh_type == SHT_REL ) {
			Elf32_Rel *relp = (Elf32_Rel *)module->section[i];
			for ( j = 0; j < module->shdr[i].sh_size/sizeof(Elf32_Rel); 
			j++ ) {
				n = relocate( module, relp++, &module->shdr[i] );
				if ( n < 0 ) {
					return( -1 );
				}
			}
		}
	}
    return( 0 );
}

unsigned long 
find_symaddr( module_t * module, const char * symbol )
{
	Elf32_Sym * sym;

	sym = localall_sym_find( module, symbol );
    if ( sym == NULL ) {
        fatal( "no symbol (%s)\n", symbol );
		return( -1 );
	}
    return( 
	(unsigned long)module->shdr[sym->st_shndx].sh_addr+sym->st_value );
}

unsigned long 
init_module( module_t * p_module, unsigned long maddr )
{
    unsigned long len;
    int i;
    int n;

    len = 0;
    p_module->maddr = maddr;
    for ( i = 0; i < p_module->ehdr.e_shnum; i++ ) {
        if ( p_module->shdr[i].sh_type != SHT_PROGBITS ) { 
            continue;
        }
        p_module->shdr[i].sh_addr = len + maddr;
        len += p_module->shdr[i].sh_size;
    }

    /* I add some operations of symbol of SHN_COMMON */
    if ( sym_index != -1 ) {
		Elf32_Sym * sym;
		int l_addr = len + p_module->maddr;

		sym = (Elf32_Sym *)p_module->section[sym_index];
		for ( i = 0; i < sym_num; i++ ) {
			if ( sym->st_shndx == SHN_COMMON ) {
				int offset;

				if ( (offset=(l_addr % sym->st_value)) ) {
					l_addr += sym->st_value - offset;
				}
				sym->st_value = l_addr;
				l_addr += sym->st_size;
			}
			sym++;
		}
		len = l_addr - p_module->maddr;
	}

    p_module->len = len;
    n = relocate_module( p_module );
    if ( n < 0 )  {
        fatal( "relocate_module failed !\n" );
		return( -1 );
	}
    return( 0 );
}

unsigned long 
make_module( module_t * module, unsigned long dst )
{
    unsigned retval = 0;
    int i;

    for ( i = 0; i < module->ehdr.e_shnum; i++ ) { 
		if ( module->shdr[i].sh_type != SHT_PROGBITS ) {
			continue;
		}
		memcpy( (void *)(dst+retval), module->section[i], 
		module->shdr[i].sh_size );
		retval += module->shdr[i].sh_size;
	}
    return( retval );
}

int 
module_size( int fd )
{
    Elf32_Ehdr ehdr;
    Elf32_Shdr * shdr;
    int slen;
    int len;
    int i;
    int n;

	n = read_kmem( &ehdr, 0, sizeof(Elf32_Ehdr), fd );
    if ( n < 0 ) {
		return( -1 );
	}

    slen = sizeof( Elf32_Shdr ) * ehdr.e_shnum;
    shdr = ( Elf32_Shdr * )malloc( slen );
    if ( shdr == NULL ) {
        perror( "malloc()" );
        return( -1 );
    }

    n = read_kmem( shdr, ehdr.e_shoff, slen, fd );
    if ( n < 0 ) {
        return( -1 );
    }

    if ( ehdr.e_type != ET_REL ) {
        fatal( "bad module..\n" );
    	return( -1 );
	}

    for ( len = i = 0; i < ehdr.e_shnum; i++ ) {
        if ( shdr[i].sh_type != SHT_PROGBITS ) {
            continue;
        }
        len += shdr[i].sh_size;
    }

    free( shdr );
    return( len );
}

int 
print_symaddr( module_t * p_module, const char * symbol )
{
    Elf32_Sym * sym;
    
    sym = localall_sym_find( p_module, symbol );
    if ( sym == NULL ) {
        return( -1 );
    }
    fprintf( stdout, "%s at %#x\n", symbol, 
	(unsigned int)(p_module->maddr+sym->st_value) );
    return( 0 );
}

void * 
rel_module( char * filename, unsigned long addr, int * len, char ** sym, 
char * function, size_t * func_addr )
{
	module_t module;
	char * p_module;
	int fd;
	int n;

    memset( &module, 0, sizeof(module_t) );
	fd = open( filename, O_RDONLY );
    if ( fd < 0 ) {
		fatal( "%s : No such module\n", filename );
		return( NULL );
	}

	n = load_module( &module, fd );
    if ( n < 0 ) {
		close( fd );
		return( NULL );
	}

	n = init_module( &module, addr );
    if ( n < 0 ) {
		close( fd );
		return( NULL );
	}

    p_module = xmalloc((*len=module_size(fd)));
    if ( !*len ) {
		close( fd );
		return( NULL );
	}

    make_module( &module, (unsigned long)p_module );
    close( fd );
	
    if ( sym != 0 ) {
        for ( fd = 0; sym[fd]; fd++ ) {
            print_symaddr( &module, sym[fd] );
		}
	}
    if ( function ) {
        *func_addr = (size_t)find_symaddr( &module, function );
	}
    sym_index = -1;
    sym_num = -1;
    return( p_module );
}

void
get_size_module( char * module )
{
	char * mod;
	char * p_mod;
	int fd_module;
	int len;

	fd_module = open( module, O_RDONLY );
	if ( fd_module < 0 ) {
		perror( "open()" );
		fputs( "\n", stdout );
		return;
	}

	len = module_size( fd_module );	
	if ( len < 0 ) {
		close( fd_module );
		return;
	}

	mod = (char *)strtok( module, "/" );
	if ( mod != NULL ) {
		while ( (mod = (char *)strtok(NULL, "/")) != NULL ) {
			p_mod = mod;
			continue;
		}
	}
	fprintf( stdout, "%s size: %d octets\n\n", p_mod, len );
	close( fd_module );
	return;
}

void
get_list_module( void )
{
	struct module_stat * m;
	int n;
	int i;

	n = get_kernel_info( K_INFO | K_SYMBOLS );
	if ( n < 0 ) {
		return;
	}

	m = module_stat;
	if ( n_module_stat > 0 ) {
		fputs( "Module\t\t\tAddress\n", stdout );
		for ( i = 0; i < n_module_stat; ++i, ++m ) {
			fprintf( stdout, "[%s]\t\t", m->name );
			fprintf( stdout, "0x%0*lx\n", (int)(2 * sizeof(void *)),
			m->addr );
		}
		fputs( "\n", stdout );
		return;
	}

	fputs( "No module installed..\n\n", stdout );
	return;
}

void 
get_info_module( char * module )
{
	struct module_stat * m;
	struct module_symbol * s;
	struct module stealth;
	char name[ BUFSIZ ];
	size_t module_addr;
	int n;
	int i;
	int j;

	n = get_kernel_info( K_INFO | K_SYMBOLS );
	if ( n < 0 ) {
		return;
	}

	m = module_stat;
	module_addr = 0;
	if ( n_module_stat > 0 ) {
		for ( i = 0; i < n_module_stat; i++, m++ ) {
			if ( !strncmp(module, m->name, strlen(module)) ) {
				module_addr = m->addr;
			}
		}
	}
			
	if ( module_addr == 0 ) {
		fatal( "module name invalid..\n" );
		return;
	}

	n = read_kmem( &stealth, module_addr, sizeof(struct module), 
	file_descriptor_kmem );
	if ( n < 0 ) { 
		return;
	}

	n = read_kmem( &name, (size_t)stealth.name, BUFSIZ, 
	file_descriptor_kmem );
	if ( n < 0 ) {
		return;
	}

	fprintf( stdout, "Probing memory at 0x%08x\n\n", module_addr );
	if ( !isprintable(name) ) {
		fatal( "address or module name invalid.\n" );
		return;
	}

	fprintf( stdout, " Name : %s\n", name );
	fputs( " Path : ", stdout );
	n = probepath( file_descriptor_kmem, module_addr, stealth );
	if ( n < 0 ) {
		return;
	}
	fputs( "\n", stdout );

	fprintf( stdout, " Size : %ld pages\n", stealth.size );
	fputs( "Flags : ", stdout );
	if ( stealth.flags & MOD_UNINITIALIZED ) {
		fputs( "MOD_UNINITIALIZED ", stdout );
	}
	if ( stealth.flags & MOD_RUNNING ) {
		fputs( "MOD_RUNNING ", stdout );
	}
	if ( stealth.flags & MOD_DELETED ) {
		fputs( "MOD_DELETED ", stdout );
	}
	if ( stealth.flags & MOD_AUTOCLEAN ) {
		fputs( "MOD_AUTOCLEAN ", stdout );
	}
	if ( stealth.flags & MOD_VISITED ) {
		fputs( "MOD_VISITED ", stdout );
	}
	if ( stealth.flags & MOD_USED_ONCE ) {
		fputs( "MOD_USED_ONCE ", stdout );
	}
	if ( stealth.flags & MOD_JUST_FREED ) {
        	fputs( "MOD_JUST_FREED ", stdout );
	}
	fputs( "\n\n", stdout );

	m = module_stat;
	fputs( "Address\t\tSymbol\n", stdout );
	for ( i = 0; i < n_module_stat; ++i, ++m ) {
		if ( !strncmp(module, m->name, strlen(module)) ) {
			for ( j = 0, s = m->syms; j < m->nsyms; 
			++j, ++s ) {
				fprintf( stdout, "0x%0*lx\t%-32s\n",
				(int)(2 * sizeof(void *)),
				s->value,
				(char *)s->name );
			}
		}
	}

	fputs( "\n", stdout );
	return;
}

#define INSMOD "/sbin/insmod"
#define RMMOD "/sbin/rmmod"
void
load_unload_module( char * module, char * operation )
{
	FILE * pipe;
	char command[ BUFSIZ ];
	char * output;

	memset( command, '\0', sizeof(command)-1 );
	snprintf( command, sizeof(command), "%s %s", operation, module );

	pipe = popen( command, "r" );
	if ( pipe == NULL ) {
		fatal( "%s don't exist..\n", operation );
		pclose( pipe );
		return;
	}

	fscanf( pipe, "%s", output );
	if ( output != NULL ) {
		fprintf( stdout, "%s\n", output );
	}
	pclose( pipe );
	return;
}

void
usage_module( void )
{
    fputs( 
	"module [ -sirdl ] <arguments>\n\n"
	"This command makes it possible to manage modules: installation,\n" 
	"suppression, informations...\n\n"
	"__options\n"
	"  -s [module name] : display module size ;\n"
	"  -l : display modules loaded ;\n"
	"  -d [module name] : display characteristics of a module ;\n"
	"  -i [module name] : load a module (pipe with /sbin/insmod) ;\n"
	"  -r [module name] : unload a module (pipe with /sbin/rmmod) ;\n\n"
	, stdout);
    return;
}

#define MODULES_NOT_SUPPORTED( modules_not_supported ) { \
	if ( modules_not_supported > 0 ) { \
		fatal( "sorry but modules aren't supported " \
		"in your system..\n\n" ); \
		return; \
	} \
}

void
do_module( int pargc, char * pargv[] )
{
	char c;

	NOARGUMENTS( pargc, 1, module );

	optind = 0;
	while ( (c = getopt(pargc, pargv, "li:r:d:s:")) != EOF ) {
		switch ( c ) {
			case 's':
				get_size_module( optarg );
				break;
			case 'i':
				MODULES_NOT_SUPPORTED( modules_not_supported ); 
				load_unload_module( optarg, INSMOD );
				break;
			case 'r':
				MODULES_NOT_SUPPORTED( modules_not_supported ); 
				load_unload_module( optarg, RMMOD );
				break;
			case 'd':
				MODULES_NOT_SUPPORTED( modules_not_supported ); 
				get_info_module( optarg );
				break;
			case 'l':
				MODULES_NOT_SUPPORTED( modules_not_supported ); 
				get_list_module();
				break;
			default:
				usage_module();
				return;
		}
	}
	return;
}

