/*
 * Copyright (C) 2002 sauron@kernsh.org
 * Copyright (C) 2002 zorgon@kernsh.org
 *
 * THIS LICENCE IS ISSUED FROM NMAP LICENCE AND MODIFIED
 *
 * The program kernsh is (C) 2002 sauron/zorgon. This program is
 * free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free
 * Software Foundation; Version 2. This guarantees your right to use,
 * modify, and redistribute this software under certain conditions.
 *
 * Source is provided to this software because we believe users have a
 * right to know exactly what a program is going to do before they run
 * it.
 *
 * Source code also allows you to fix bugs, and add new features.
 * You are highly encouraged to send your changes to dev@kernsh.org for
 * possible incorporation into the main distribution. By sending these
 * changes to sauron@kernsh.org or zorgon@kernsh.org, it is assumed that
 * you are offering them the unlimited, non-exclusive right to reuse,
 * modify, and relicense the code. This is important because the
 * inability to relicense code has caused devastating problems for other
 * Free Software projects (such as KDE and NASM). Kernsh will always be
 * available Open Source. If you wish to specify special license
 * conditions of your contributions, just say so when you send them.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details (
 * http://www.gnu.org/copyleft/gpl.html ).
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <linux/module.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/utsname.h>

#define SYS_HOOKED	250
#define MAX_PARAM	10
#define DEFAULT_FILE	"client_struct.rc"
#define DEFINE		"#define"
#define ERR		"E: "
#define VERT		"\033[32m"
#define NORM		"\033[0m"
#define F(x)            { fprintf(stderr, ERR"%s", x); exit(-1); }
#define P(x)            { perror(x); exit(-1);}

#define T_INT       'i'
#define T_CHAR      'c'
#define T_STRING    's'
#define T_POINTER   'p'

struct new_module_symbol
{
  void * value;
  unsigned long name;
};

typedef struct param
{
  int	pid;
  char	*str;
} parm_t;

typedef struct cmd
{
  void	(*func)(parm_t *);
  char	*name, *description;
} cmd_t;

typedef struct s_exec {
    int ret;
    char fmt[MAX_PARAM];
    char chartab[MAX_PARAM];
    short shorttab[MAX_PARAM];
    char * stringtab[MAX_PARAM];
    unsigned long u32tab[MAX_PARAM];
    void * function;
    int nbparam;
} t_exec;

typedef struct define_s
{
  char	*name;
  char	type;
  int	val;
  char	val2;
  char	*val3;
  struct define_s	*next;
} t_def;

enum
{
  TYPE_CHAR,
  TYPE_STRING,
  TYPE_INT,
  TYPE_POINTEUR
};

int query_module(const char *name, int which,
                 void *buf, size_t bufsize, size_t *ret);

char	**word_tab(char *, char);
int	vsnprintf (char *, size_t, const char *, va_list);
void	err(char *, ...);
void	help(parm_t *);
void	parse_call(parm_t *);
int	ishexa(char *);
int	isnumber(char *);
void    *xmalloc(int);
void	Exit(parm_t *);
void	clientstruct_main(parm_t *);
char	*strndup(char *, int);
int	isdefined(char *);
char	print_define(char *, t_exec *, int);
char	*delspace(char *);
char	*mmapfile(char *);
char	*handlemagic(char *);
void	*get_addr(char *);

t_def	* list = 0;
cmd_t cmdtab[] = {
  {parse_call, "call", "call printk(\"toto %d\\n\", 42)"},
  {Exit, "exit", "Quitter"},
  {Exit, "quit", "Quitter"},
  {help, "help", "Afficher ce menu.."},
  {0,0}};

void
launch_exec( t_exec * list, char *funcname)
{
    long __res;

    list->ret = -1;
    __asm__ volatile ( "int $0x80"
                : "=a" (__res)
                : "0" (SYS_HOOKED),"b" ((long)(list)) );
    if (list->ret != -1)
      printf("%s return : %#x(%d)\n", funcname, list->ret, list->ret);
}

char *    getversion()
{
  struct utsname *s;

  s = malloc(sizeof(struct utsname));
  uname(s);
  return (s->release);
}

char *	foundmap()
{
  int len;
  char * buf, *data;
  char *version;

  version = getversion();
  len = strlen(version) + strlen("/boot/System.map-") + 1;
  buf = malloc (len);
  bzero(buf, len);
  strcpy (buf, "/boot/System.map-");
  strcat (buf, version);
  if (!(data = mmapfile(buf)))
    if (!(data = mmapfile("/boot/System.map")))
      if (!(data = mmapfile("/System.map")))
	if(!(data = mmapfile("/usr/src/linux/System.map")))
	  P("System.map not found !");
  free(buf);
  return (data);
}

char * mmapfile(char *filename)
{
  int           fd;
  char          *map;
  struct stat   info;

  if((fd = open(filename, O_RDONLY)) < 0)
    return (0);
  if ((stat(filename, &info) == -1))
    return (0);
  if((map = mmap (0, info.st_size, PROT_READ, MAP_SHARED, fd, 0)) == (char *) MAP_FAILED)
    return (0);
  return (map);
}

u_long  get_sym_addr(char **map, char *symbol)
{
  int i;
  char *faya;
  u_long value;
  
  if (get_addr(symbol))
    return ((u_long) get_addr(symbol));
  for (i = 0; map[i]; i++)
    {
      printf("comparation entre %s et [%s]\n", map[i] + 11, symbol);
      if (strlen(map[i]) == strlen(symbol) + 11)
	if (!strcmp(map[i] + 11, symbol))
	  {
	    faya = xmalloc(11);
	    strcpy(faya, "0x");
	    strncat(faya, map[i], 8);
	    value = strtoul(faya, 0, 0);
	    if (!value)
	      P("Cannot found symbol..");
	    free(faya);
	    return (value);
	  }
    }
  return (0);
}

void * get_addr(char *sym)
{
  struct new_module_symbol *syms, *s;
  size_t ret, bufsize, j;

  (u_long) syms = malloc(bufsize = sizeof(struct new_module_symbol));
  retry :
    if ((query_module(NULL, 4, syms, bufsize, &ret)) == -1)
      {
	if (errno == ENOSPC)
	  {
	    syms = (struct new_module_symbol *) realloc(syms, bufsize = ret);
	    goto retry;
	  }
	else
	  return (0);
      }

  for (j = 0, s = syms; j < ret; ++j, ++s)
    {
      if(strstr((char *)syms+s->name, sym))
	{
	  free(syms);
	  return (s->value);
	}
    }
  free(syms);
  return (0);
}

void    *xmalloc(int size)
{
  char  *ret;

  if (!(ret = malloc(size)))
    F("malloc");
  for (size--; size >= 0; size--)
    ret[size] = '\0'; 
  return (ret);
}

void err(char *fmt, ...)
{
  char buf[256];
  va_list ap;

  va_start(ap, fmt);
  vsnprintf(buf, 256, fmt, ap);
  buf[255] = '\0';
  printf("E: %s\n", buf);
  va_end(ap);
  exit (-1);
}

int	isdefined(char *str)
{
  t_def *tmp;

  for (tmp = list; tmp; tmp = tmp->next)
    {
      if (strlen(str) == strlen(tmp->name))
	if (!strcmp(tmp->name, str))
	  return (1);
    }
  return (0);
}

char	print_define(char *str, t_exec * cmd, int i)
{
  t_def *tmp;

  for (tmp = list; tmp; tmp = tmp->next)
    {
      if (strlen(str) == strlen(tmp->name))
	if (!strcmp(tmp->name, str))  
	  switch(tmp->type)
	    {
	    case TYPE_CHAR:
	      cmd->chartab[i] = tmp->val2;
	      return (T_CHAR);
	      break;
	    case TYPE_POINTEUR:
	      cmd->u32tab[i] = tmp->val;
	      return (T_POINTER);
	      break;
	    case TYPE_INT:
	      cmd->u32tab[i] = tmp->val;
	      return (T_INT);
	      break;
	    case TYPE_STRING:
	      cmd->stringtab[i] = tmp->val3;
	      return (T_STRING);
	      break;
	    default:
	      break;
	    }
    }
  return ('u');
}

int	additem(char *name, char type, char val, int val2, char *val3)
{
  t_def * new;

  new = malloc(sizeof(t_def));
  new->next = list;
  new->name = strdup(name);
  new->type = type;
  switch(type)
    {
    case TYPE_CHAR:
      new->val2 = val;
      break;
    case TYPE_POINTEUR:
    case TYPE_INT:
      new->val = val2;
      break;
    case TYPE_STRING:
      new->val3 = val3;
      break;
    default:
      free(new);
      return (0);
    }
  list = new;
  return (1);
}

char	*get_string(char *str)
{
  return (handlemagic(strndup(str + 1, strlen(str + 1) - 1)));
}

char	get_char(char *str)
{
  if (strlen(str) == 3)
    return (*(str + 1));
  if (*(str + 2) == 'n')
    return ('\n');
  if (*(str + 2) == 't')
    return ('\t');
  return (0);
}

int	parsedef(char *name, char *str)
{
  if (strlen(str) > 2)
    if (!strncmp(str, "0x", 2) && ishexa(str + 2) && strlen(str + 2) <= 8)
      return(additem(name, TYPE_POINTEUR, 0, strtoul(str, 0, 0), 0));
  if (str[0] == '"' && str[strlen(str) - 1] == '"')
    return(additem(name, TYPE_STRING, 0, 0, get_string(str)));
  if (str[0] == '\'' && str[strlen(str) - 1] == '\'' && ((strlen(str) == 3) || str[1] == '\\'))
    return(additem(name, TYPE_CHAR, get_char(str), 0, 0));
  if (isnumber(str))
    return(additem(name, TYPE_INT, 0, atoi(str), 0));
  return (0);
}

int	get_len(char *str)
{
  int i;

  for (i = 0; str[i]; i++)
    if (str[i] == ' ' || str[i] == '\t' || str[i] == '\n')
      return (i);
  return (i);
}

int	parseline(char * line)
{
  int nbn, nbv, ret, len = strlen(DEFINE);
  char *name;
  char *value;

   if (strlen(line) <= len)
    return (0);
   if (strncasecmp(line, DEFINE, len))
    return (0);
  nbn = get_len(delspace(line + len));
  name = strndup(delspace(line + len), nbn);
  nbv = get_len(delspace(delspace(line + len) + nbn));
  value = strndup(delspace(delspace(line + len) + nbn), nbv);
  ret = parsedef(name, value);
  free(name);
  free(value);
  return (ret);
}

void	parseconf(void)
{
  char **map, *tmp;
  int i;

  if(!(tmp = mmapfile(DEFAULT_FILE)))
     return;
  map = word_tab(tmp, '\n');
  for (i = 0; map[i]; i++)
    if (map[i][0])
      parseline(map[i]);
}

int	get_first_occur(char *str, char c)
{
  int i;

  for (i = 0; str[i]; i++)
    if (str[i] == c)
      return (i);
  return (-1);
}

char	*strndup(char *str, int len)
{
  return (strncpy(xmalloc(len + 1), str, len));
}

#define MAX_PARAM	10

char	*delspace(char *str)
{
  do
    if (*str != ' ' && *str != '\t')
      return (str);
  while (*(str++));
  return (str);
}

char **word_tab(char *str, char c)
{
  char **tab;
  int i;
  int j;
  int u;
  int k;

  if (!*str)
    return (0);

  for (j = i = 0; str[j]; j++)
    if (str[j] == c)
      i++;
  tab = malloc((i + 2) * sizeof(char *));
  i = k = j = u = 0;
  do
    {
      u = str[i] == '"' ? u ? 0 : 1 : str[i] == '\'' ? u ? 0 : 1 : u;
      if(str[i] == c && !u)
	{
	  tab[j++] = delspace(strndup(&str[k], i - k));
	  k = i + 1;
	}
      i++;
    }
  while(str[i]);
  tab[j++] = delspace(&str[k]);
  tab[j] = 0;
  return (tab);
}

int	isnumber(char *str)
{
  int	i;

  if (!str || !*str)
    return (0);
  for (i = 0; str[i]; i++)
    if (str[i] < '0' || str[i] > '9')
      return (0);
  return (1);
}

int	ishexa(char *str)
{
  int	i;

  if (!str || !*str)
    return (0);
  for (i = 0; str[i]; i++)
    if (!((str[i] >= '0' && str[i] <= '9') 
	|| ((str[i] >= 'a' && str[i] <= 'z')) 
	|| ((str[i] >= 'A' && str[i] <= 'Z'))))
      return (0);
  return (1);
}

char	*handlemagic(char *str)
{
  int i;
  
  for (i = 0; str[i]; i++)
    if (str[i] == '\\')
	switch(str[i + 1])
	    {
	    case '\\':
	      i++;
	      break;
	    default:
	      if((str[i] = str[i + 1] == 'n' ? '\n' : str[i + 1] == 't' ? '\t' : str[i + 1] == 'r' ? '\r' : 0))
		strcpy(str + i + 1, str + i + 2);
	      break;
	    }
  return (str);
}

void	parse(char *str)
{
  int i, j;
  char *function, **param;
  t_exec	cmd;
  char **map = word_tab(foundmap(), '\n');

  if((i = get_first_occur(str, '(')) == -1)
    return;
  if((j = get_first_occur(str, ')')) == -1)
    return;
  function = strndup(str, i);
  param = word_tab(strndup(str + i + 1, j - i - 1), ',');
  memset(&cmd, 0, sizeof(t_exec));
  for (i = 0; param[i] && i <= MAX_PARAM; i++)
    {
      if (strlen(param[i]) > 2)
	if (!strncmp(param[i], "0x", 2) && ishexa(param[i] + 2) && strlen(param[i] + 2) <= 8)
	  {
	    cmd.u32tab[i] = strtoul(param[i], 0, 0);
	    cmd.fmt[i] = T_POINTER;
	    continue;
	  }
      if (param[i][0] == '"' && param[i][strlen(param[i]) - 1] == '"')
	{
	  cmd.stringtab[i] = handlemagic(strndup(param[i] + 1, strlen(param[i] + 1) - 1));
	  cmd.fmt[i] = T_STRING;
	  continue;
	}
      if (param[i][0] == '\'' && param[i][strlen(param[i]) - 1] == '\'' && ((strlen(param[i]) == 3) || param[i][1] == '\\'))
	{
	  cmd.chartab[i] = get_char(param[i]);
	  cmd.fmt[i] = T_CHAR;
	  continue;
	}
      if (isnumber(param[i]))
	{
	  cmd.u32tab[i] = atoi(param[i]);
	  cmd.fmt[i] = T_INT;
	  continue;
	}
      if (isdefined(param[i]))
	{
	  cmd.fmt[i] = print_define(param[i], &cmd, i);
	  continue;
	}
      printf("parse error\n");
      return;
    }
  if ((cmd.function = (void *) get_sym_addr(map, delspace(function))) == 0)
    printf(ERR"function %s not found\n", function);
  else
    {
      cmd.nbparam = i;
      launch_exec(&cmd, delspace(function));
    }
}

void	err_handler(int num, int ret, char *msg)
{
  if (ret == num)
    err("%s", msg);
}

void	help(parm_t *elem)
{
  int	i;

  printf("-- command list : --\n");
  for (i = 0; cmdtab[i].name; i++)
    printf("%s:\t%s\n", cmdtab[i].name, cmdtab[i].description);
  printf("-- end of command list --\n");
}

void	parse_call(parm_t *elem)
{
  parse(elem->str);
}

void	Exit(parm_t *elem)
{
  printf("client_struct exiting...\n");
  exit(0);
}

#define	BUF_SIZE	512

void	clientstruct_main(parm_t *elem)
{
  char	buffer[BUF_SIZE + 1];
  int	i, j;

	fputs( "[!] client_struct launched !! You have to be sure that\n"
	"[!] module is inserted and that syscall 250 is redirect on the\n"
	"[!] function our_handler of struct_lkm !\n", stdout );
  while (1)
    {
      memset(buffer, 0, BUF_SIZE + 1);
      printf(VERT"clienstruct"NORM"%% ");
      fflush(stdout);
      read(0, buffer, BUF_SIZE);
      j = -1;
      for (i = 0; cmdtab[i].name; i++)
	if (strlen(buffer) >= strlen(cmdtab[i].name))
	  if (!strncmp(cmdtab[i].name, buffer, strlen(cmdtab[i].name)))
	    {
	      elem->str = buffer + strlen(cmdtab[i].name);
	      cmdtab[i].func(elem);
	      j++;
	      break;
	    }
      if (j == -1)
	  printf("%s", *buffer == '\n' ? "" : "clientstruct: Invalid command\n");
    }
}

int	main(int ac, char **av)
{
  parm_t	elem;

  parseconf();
  clientstruct_main(&elem);
  return (0);
}




